#The Serial GRBL class handles all levels of comunications to the GRBL 
#controller
#Todo: 
#-error lock, stop sending code if errors occur
#-add get and set settings
#-check settings to list of required settings
#-auto set settings if different
#-Get coordinates from status
#-return coordinates and status from function

import serial
import threading
import time

class GRBL(serial.Serial):
    def __init__(self):
        self.ser = serial.Serial() #make an instance of serial connection
        self.ser.baudrate = 115200 #set baudrate
        self.ser.timeout = 0 #set timeout to 0, non blocking mode
        
        #status flags
        self.connection_state = 0 #whether connected or not
        self.started_state = 0 
        self.ok_state = 1 #whether the response of serial has been OK or not
        self.error_state = 0 #error of system
        self.motion_state = "" #what the GRBL is doing (0 is idle
        self.homed_state = 0 #if printer is homed
        
        self.send_get_status = 0 #whether to send ?
        self.motion_x_pos = 0.0
        self.motion_y_pos = 0.0
        self.motion_z_pos = 0.0
        
        self.printer_size_x = 380.0 #x total size
        self.printer_size_y = 250.0 #y total size
        self.printer_size_f = 100.0 #feed hopper total size
        self.printer_size_b = 100.0 #build hopper total size
        
        self.printer_homing_dir_x = -1 #homing direction
        self.printer_homing_dir_y = 1 #homing direction
        self.printer_homing_pulloff = 5 #how much the printer steps back after a home
        
        self.window_output_buffer = "" #holds a buffer of what was sent out
        self.window_input_buffer = "" #holds a buffer of what was received
        
        self.grbl_version = 0.0 #version of grbl
        
        self.gcode_buffer = []
        self.gcode_buffer_left = 0
        
    
    def Connect(self, serial_port):
        """Attempt to connect to the GRBL controller"""
        self.com_port_raw = str(serial_port) #get value from set_com
        self.ser.port = self.com_port_raw #set com port .
        
        if (self.connection_state == 0): #if not yet connected
            #print("attempting to open: " + self.com_port_raw)
            self.temp_com_success = 0
            try: #try to open a com port with it
                self.ser.open()
                self.temp_com_success = 1
            except:
                #print ("Unable to open connection")
                nothing = 0
            if (self.temp_com_success == 1):
                print(self.com_port_raw + " for grbl opened")
                self.connection_state = 1
                self.started_state = 0 
                self.ok_state = 1 
                self.error_state = 0 
                self.homed_state = 0 
                self._stop_event = threading.Event()
                self.update_thread = threading.Thread(target=self.Update)
                self.update_thread.start()
                self.status_thread = threading.Thread(target=self.GetStatus)
                self.status_thread.start()
                return 1
            else:
                return 0
                
    def Disconnect(self):
        """close the connection to GRBL"""
        if (self.connection_state == 1):
            self._stop_event.set()
            self.ser.close()
            print("Closing grbl connection")
            self.connection_state = 0
            return 0
        
    def Update(self):
        """Preforms all continous tasts required for GRBL connection"""
        time.sleep(0.05)
        read_buffer = "" #used to store Serial read data
        while not self._stop_event.is_set():
            #get code
            temp_success = 0 #success value
            try: #attempt to read serial
                if (self.ser.in_waiting > 0):
                    temp_read = self.ser.read(self.ser.in_waiting) #add serial to read buffer
                    temp_read = str(temp_read.decode('utf-8'))
                    temp_success = 1
            except:
                print("Read error") #some mistake, otherwise ignore quietly
                break
            if temp_success == 1: 
                read_buffer += temp_read
                #print(read_buffer)
                #add line to read buffer
            
            temp_decode = read_buffer.partition('\n') #check for EOL conditions
            if (temp_decode[1] == '\n'): #if '\n' 
                read_buffer = temp_decode[2] #write remainder to buffer
                read_line = str(temp_decode[0])
                read_line = read_line.lower() #make all lower case for checking 
                read_line = read_line.rstrip() #remove carriage return
                #print("reading line: " + str(read_line)) 
                #check purpose of response
                if (read_line.startswith('ok')): #if ok was found,
                    self.ok_state = 1 #set ok state to 1
                    if (self.homed_state == 2):
                        self.homed_state = 1 #first OK after homing is homing done, set to 1
                        #print("Is homed")
                    #print("OK found, setting ok state")
                if (read_line.startswith('error')): #if error was found,
                    self.window_input_buffer += str(read_line) #add to the window buffer
                    self.error_state = 1
                    print("error found, getting error type")
                if (read_line.startswith('[')): #if message was found,
                    self.window_input_buffer += str(read_line) #add to the window buffer
                    print("message found, getting message")
                    #self.Home()
                if (read_line.startswith('<')): #if status was found,
                    #print("status found, getting status")
                    read_line = read_line.partition('|')#partition till |, compare options
                    temp_line = read_line[0].lstrip('<') #get value
                    read_line = read_line[2] #set remainder
                    self.motion_state = str(temp_line) #set status
                    read_line = read_line.partition(':')
                    temp_line = read_line[0]
                    read_line = read_line[2]
                    if (temp_line == "wpos"):
                        read_line = read_line.partition(',')
                        temp_line = read_line[0] #get x
                        self.motion_x_pos = float(temp_line)
                        read_line = read_line[2]
                        
                        read_line = read_line.partition(',')
                        temp_line = read_line[0] #get y
                        self.motion_y_pos = float(temp_line)
                        read_line = read_line[2]
                        
                        read_line = read_line.partition('|')
                        temp_line = read_line[0] #get z
                        self.motion_z_pos = float(temp_line)
                        read_line = read_line[2]
                    
                    #print("Status: " + self.motion_state)
                if (read_line.startswith('grbl')): #if grbl was found,
                    #print("grbl found, getting version")
                    read_line = read_line.partition(' ') #strip grbl from file
                    read_line = read_line[2].partition('f ') #strip instructions past version number
                    if (read_line[1] == 'f '):
                        self.window_input_buffer += str(read_line) #add to the window buffer
                        self.grbl_version = float(read_line[0])
                        print("GRBL Version: " + str(self.grbl_version))
                        self.started_state = 1
                        
                
            #is ok state is 1, and line buffered, send new line
            if (self.ok_state == 1 and self.started_state == 1):
                if (self.send_get_status == 1): 
                    self.ok_state = 0 #set ok state to 0
                    self.SerialWriteRaw("?\r",0) #send ?
                    self.send_get_status = 0 #set get status to 0
                    #print("Getting status")
                elif (self.BufferLeft() > 0): #if there are lines left to print
                    self.ok_state = 0 #set ok state to 0
                    self.BufferNext() #print next line in buffer to serial
        
    def SerialWriteRaw(self, input_string, temp_priority):
        """prints a line to the GRBL (no checks)
        priority is 0 for not send to output, and 1 for sent to output"""
        if (temp_priority == 1):
            self.window_output_buffer += input_string #add to the window buffer
        self.ser.write(input_string.encode('utf-8'))
        
    def SerialWriteBufferRaw(self, input_string):
        """Adds a line to the input buffer""" 
        if (self.connection_state == 1): #only work when connected
            self.gcode_buffer.append(str(input_string) + '\r') #add string to buffer
            self.gcode_buffer_left += 1 #add one to left value
    
    def SerialGotoXY(self, temp_x, temp_y, temp_f=''):
        """move to the given X/Y position at feedrate in mm/min"""
        temp_string = "G1 X"
        temp_string += str(temp_x)
        temp_string += " Y"
        temp_string += str(temp_y)
        if (temp_f):
            temp_string += " F"
            temp_string += str(temp_f)
        self.SerialWriteBufferRaw(temp_string)
        
    def Jog(self, temp_axis, temp_distance, temp_feed):
        """Jogs the given axis for the given distance at the given feed,
        Has no protections built in"""
        self.SerialWriteBufferRaw("G91") #set printer to relative (G91)
        #move axis relatively
        temp_string = "G1 "
        temp_string += str(temp_axis)
        temp_string += str(temp_distance)
        temp_string += " F"
        temp_string += str(temp_feed)
        self.SerialWriteBufferRaw(temp_string)
        self.SerialWriteBufferRaw("G90")#set printer to absolute (G90)
        
    def BufferLeft(self):
        """returns how many lines are left in the buffer"""
        return self.gcode_buffer_left
    
    def BufferNext(self):
        """Writes the next line in the buffer to the serial"""
        if (self.BufferLeft() > 0): #if there are lines left in the buffer
            self.gcode_buffer_left -= 1 #subtract 1 from left value
            self.SerialWriteRaw(self.gcode_buffer[0],1) #print to GRBL
            del self.gcode_buffer[0] #remove the written line
    
    def Home(self):
        """Homes the printer and sets the coordinates
        Because homing only sends the ok after the home, it suffices
        to simply send the move to home ('$H') and then 
        G92 X# Y#. The ok buffer will take care of the rest"""
        self.SerialWriteBufferRaw('$h') #buffer home command
        #calculate x and y home based on home dir and length
        if (self.printer_homing_dir_x == -1):
            temp_x = self.printer_homing_pulloff
        else:
            temp_x = self.printer_size_x -self.printer_homing_pulloff
        
        if (self.printer_homing_dir_y == -1):
            temp_y = self.printer_homing_pulloff
        else:
            temp_y = self.printer_size_y -self.printer_homing_pulloff

        self.SerialWriteBufferRaw('G92 X' + str(temp_x) + ' Y' + str(temp_y))#buffer G92 command
        self.homed_state = 2 #set homed state to 2 (homing in progress)
    
    def GetStatus(self):
        """periodically sends a ? to get status"""
        time.sleep(5) #initial wait to get system time to start
        while not self._stop_event.is_set():
            time.sleep(0.5) #wait for 0.5 seconds
            self.send_get_status = 1
            
    def GetWindowOutput(self):
        """returns the entire string of what was sent since the 
        last call of this function, then clears that buffer"""
        temp_return = self.window_output_buffer #write to return value
        self.window_output_buffer = "" #clear buffer
        return temp_return #return response
        
    def GetWindowInput(self):
        """returns the entire string of what was received since the 
        last call of this function, then clears that buffer"""
        temp_return = self.window_input_buffer #write to return value
        self.window_input_buffer = "" #clear buffer
        return temp_return #return response

if __name__ == '__main__':
    printer = GRBL()
    temp_com = input("give com port input: ")
    printer.Connect(str(temp_com))
    while 1:
        #temp_out = input("Send command: ")
        #temp_out = str(temp_out)
        #printer.SerialWriteBufferRaw(temp_out)
        temp_x = input("X: ")
        temp_y = input("Y: ")
        printer.SerialGotoXY(temp_x, temp_y, 12000)
        


#messages expected:
#Grbl 1.1f ['$' for help] #initial welcome
#[MSG:'$H'|'$X' to unlock] #unlock message
#ok
#error:#
#<Idle|WPos:0.000,250.000,0.000|Bf:15,127|FS:0,0>
